import collections

import tensorflow as tf
import numpy as np
import math as m

from Neural_Networks import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

SMALL_CONSTANT = 1e-6

class VariationalAutoencoder(object):

    def __init__(self, name, n_input, n_hidden, n_weights, middle="gaussian"):
        
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_weights = n_weights
        self.name = name
        self.middle = middle
        self.bias_start = 0.0

        network_weights = self._create_weights()
        self.weights = network_weights

        self.nonlinearity = tf.nn.relu

    def calc_reconstruction(self, z):
        with tf.name_scope("VICI_decoder"):
            if self.middle == "bernoulli":
                hidden1_pre = tf.add(tf.matmul(z, self.weights['ELBO_decoder']['W1_to_hidden']), self.weights['ELBO_decoder']['b1_to_hidden'])
                hidden1_post = self.nonlinearity(hidden1_pre)
                
                hidden3_pre = tf.add(tf.matmul(hidden1_post, self.weights['ELBO_decoder']['W1c_htoh']), self.weights['ELBO_decoder']['b1c_htoh'])
                hidden3_post = self.nonlinearity(hidden3_pre)

                hidden2_pre = tf.add(tf.matmul(hidden3_post, self.weights['ELBO_decoder']['W1b_htoh']), self.weights['ELBO_decoder']['b1b_htoh'])
                hidden2_post = self.nonlinearity(hidden2_pre)

                y_pre = tf.add(tf.matmul(hidden2_post, self.weights['ELBO_decoder']['W2_to_y_pre']), self.weights['ELBO_decoder']['b2_to_y_pre'])
                y = tf.sigmoid(y_pre)
                return y
            elif self.middle == "gaussian":

                layer_1 = tf.add(tf.matmul(z,self.weights['VICI_decoder']['W_layer_1']), self.weights['VICI_decoder']['b_layer_1'])
                previous_layer_out = self.nonlinearity(layer_1)

                layer_2 = tf.add(tf.matmul(previous_layer_out,self.weights['VICI_decoder']['W_layer_2']), self.weights['VICI_decoder']['b_layer_2'])
                previous_layer_out = self.nonlinearity(layer_2)

                mu = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_decoder']['W_layer_out_mu']), self.weights['VICI_decoder']['b_layer_out_mu'])
                mu = tf.sigmoid(mu)  # see paper
                log_sigma_sq = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_decoder']['W_layer_out_log_sigma']), self.weights['VICI_decoder']['b_layer_out_log_sigma'])
                
                return mu, log_sigma_sq
            else:
                RuntimeError(self.middle + " is not yet constructed for reconstruction")


    def _create_weights(self):
        all_weights = collections.OrderedDict()

        # Decoder
        with tf.variable_scope("VICI_DEC"):
            all_weights['VICI_decoder'] = collections.OrderedDict()
            if self.middle == "gaussian":
                hidden_number_decoder = self.n_weights
 
                all_weights['VICI_decoder']['W_layer_1'] = tf.Variable(vae_utils.xavier_init(self.n_hidden, hidden_number_decoder), dtype=tf.float32)
                all_weights['VICI_decoder']['b_layer_1'] = tf.Variable(tf.zeros([hidden_number_decoder], dtype=tf.float32)  * self.bias_start)    
                
                all_weights['VICI_decoder']['W_layer_2'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, hidden_number_decoder), dtype=tf.float32)
                all_weights['VICI_decoder']['b_layer_2'] = tf.Variable(tf.zeros([hidden_number_decoder], dtype=tf.float32)  * self.bias_start)
    
                all_weights['VICI_decoder']['W_layer_out_mu'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, self.n_input), dtype=tf.float32)
                all_weights['VICI_decoder']['b_layer_out_mu'] = tf.Variable(tf.zeros([self.n_input])  * self.bias_start, dtype=tf.float32)
                
                all_weights['VICI_decoder']['W_layer_out_log_sigma'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, self.n_input), dtype=tf.float32)
                all_weights['VICI_decoder']['b_layer_out_log_sigma'] = tf.Variable(tf.zeros([self.n_input])  * self.bias_start, dtype=tf.float32)
            
            elif self.middle == "bernoulli":
                hidden_number_decoder = 200
                all_weights['decoder']['W1_to_hidden'] = tf.Variable(vae_utils.xavier_init(self.n_hidden, hidden_number_decoder))
                tf.summary.histogram("W1_to_hidden", all_weights['decoder']['W1_to_hidden'])
    
                all_weights['decoder']['W1b_htoh'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, hidden_number_decoder))
                tf.summary.histogram("W1b_htoh", all_weights['decoder']['W1_to_hidden'])
                
                all_weights['decoder']['W1c_htoh'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, hidden_number_decoder))
                tf.summary.histogram("W1c_htoh", all_weights['decoder']['W1_to_hidden'])
    
                all_weights['decoder']['W2_to_y_pre'] = tf.Variable(vae_utils.xavier_init(hidden_number_decoder, self.n_input))
                tf.summary.histogram("W2_to_y_pre", all_weights['decoder']['W1_to_hidden'])
    
                all_weights['decoder']['b1_to_hidden'] = tf.Variable(tf.ones([hidden_number_decoder], dtype=tf.float32) * self.bias_start)
                tf.summary.histogram("b1_to_hidden", all_weights['decoder']['b1_to_hidden'])
    
                all_weights['decoder']['b1b_htoh'] = tf.Variable(tf.ones([hidden_number_decoder], dtype=tf.float32) * self.bias_start)
                tf.summary.histogram("b1b_htoh", all_weights['decoder']['b1b_htoh'])
                
                all_weights['decoder']['b1c_htoh'] = tf.Variable(tf.ones([hidden_number_decoder], dtype=tf.float32) * self.bias_start)
                tf.summary.histogram("b1c_htoh", all_weights['decoder']['b1c_htoh'])
    
                all_weights['decoder']['b2_to_y_pre'] = tf.Variable(tf.ones([self.n_input], dtype=tf.float32) * self.bias_start)
                tf.summary.histogram("b2_to_y_pre", all_weights['decoder']['b2_to_y_pre'])
    
            else:
                raise RuntimeError
            
            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights
